/* reporterd
 * 
 * Copyright © 2011 Canonical Ltd.
 * Author: Evan Dandrea <evan.dandrea@canonical.com>
 * 
 * MIT license
 */

#define _XOPEN_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <gio/gio.h>
#include <assert.h>
#include <curl/curl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/file.h>
#include <errno.h>
#include <signal.h>
#include <NetworkManager/NetworkManager.h>

#include "../lib/bson/bson.h"

#define CRASH_DB_URL "http://localhost:8080"
#define NETWORK_MANAGER_SERVICE "org.freedesktop.NetworkManager"
#define NETWORK_MANAGER_OBJECT "/org/freedesktop/NetworkManager"
#define NETWORK_MANAGER_INTERFACE "org.freedesktop.NetworkManager"
#define DBUS_PROPERTIES_INTERFACE "org.freedesktop.DBus.Properties"

/* Test cases:
 * - Hundreds of crash reports; do not balloon memory massively.
 * - Tests of various busted reports. ":\n", for example.
 * - No Internet, but lots of reports. Do not balloon memory massively.
 */

/* If true, we have an active Internet connection. True by default in case we
 * can't bring up NetworkManager */
static gboolean online_state = TRUE;
/* TODO This queue could get quite large if a lot of crashes occur when there
 * isn't an Internet connection. Should we just iterate /var/crash/ *.upload
 * every so often? Or should we clear the queue when we don't have an Internet
 * connection, and rebuild it when we do?
 */
static GQueue* report_queue = NULL;
static GMainLoop* loop = NULL;

/* Options */

static gboolean run_daemonized = FALSE;

static GOptionEntry option_entries[] = {
    { "daemonize", 'd', 0, G_OPTION_ARG_NONE, &run_daemonized,
      "Run as a daemon", NULL },
    { NULL }
};

static void
parse_arguments (int* argc, char** argv[])
{
    GError* err = NULL;
    GOptionContext* context;
    context = g_option_context_new ("Reporting client");
    g_option_context_add_main_entries (context, option_entries, NULL);
    g_option_context_parse (context, argc, argv, &err);
    g_option_context_free (context);
    if (err)
        goto error;
    return;

    error:
        g_print ("reporterd: %s\n", err->message);
        g_error_free (err);
        exit (EXIT_FAILURE);
}

char*
upload_to_crash_file (const char* upload_file)
{
    /* TODO: Why are we adding crash files instead of upload files to the list?
     */
    char* crash_file = NULL;
    const char* p = NULL;
    int len, i = 0;

    /* Find the file extension. */
    p = upload_file;
    while (*p != '\0') {
        p++; i++;
        if (*p == '.')
            len = i;
    }
    crash_file = malloc (len + 7); /* .crash */
    memcpy (crash_file, upload_file, len);
    strcpy (crash_file + len, ".crash");
    return crash_file;
}

void
append_key_value (gpointer key, gpointer value, gpointer bson_string)
{
    /* Takes a key and its value from a GHashTable and adds it to a BSON string
     * as key and its string value. */

    /* Skip the core dump for now, but we'll want to send it in a second pass
     * as the raw base64 string. */
    if (!strcmp ("CoreDump", key))
        return;
    bson_append_string ((bson*)bson_string, (char*)key, (char*)value);
}

size_t
server_response (char* ptr, size_t size, size_t nmemb, void** user_data)
{
    char* buffer = NULL;

    buffer = malloc (size * nmemb + 1);
    memcpy (buffer, ptr, size * nmemb);
    buffer[size * nmemb] = '\0';
    *user_data = buffer;
    return size * nmemb;
}

void
split_string (char* head, char** tail)
{
    *tail = strchr (head, ' ');
    if (*tail) {
        **tail = '\0';
        (*tail)++;
    }
}

void
bsonify (GHashTable* report, bson* b, const char** bson_message,
         int* bson_message_len)
{
    assert (report != NULL);

    bson_init (b);
    g_hash_table_foreach (report, append_key_value, b);
    bson_finish (b);

    printf ("Report:\n");
    bson_print (b);
    *bson_message = bson_data (b);
    *bson_message_len = bson_size (b);
}

gboolean
upload_report (const char* message_data, int message_len)
{
    CURL* curl = NULL;
    CURLcode result_code = 0;
    long response_code = 0;
    struct curl_slist* list = NULL;
    char* response_data = NULL;
    char* command = NULL;

    /* TODO use curl_share for DNS caching. */
    /* TODO use CURLOPT_READFUNCTION to transparently compress data with
     * Snappy. */
    /* TODO initialize elsewhere. */
    curl_global_init (CURL_GLOBAL_ALL);
    curl = curl_easy_init ();
    curl_easy_setopt (curl, CURLOPT_POST, 1);
    curl_easy_setopt (curl, CURLOPT_NOPROGRESS, 1);
    list = curl_slist_append (list, "Content-Type: application/octet-stream");
    curl_easy_setopt (curl, CURLOPT_URL, CRASH_DB_URL);
    curl_easy_setopt (curl, CURLOPT_HTTPHEADER, list);
    curl_easy_setopt (curl, CURLOPT_POSTFIELDSIZE, message_len);
    curl_easy_setopt (curl, CURLOPT_POSTFIELDS, (void*)message_data);
    curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, server_response);
    curl_easy_setopt (curl, CURLOPT_WRITEDATA, &response_data);
    curl_easy_setopt (curl, CURLOPT_VERBOSE, NULL);

    result_code = curl_easy_perform (curl);
    curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &response_code);

    printf ("Sent; server replied with: %s\n",
        curl_easy_strerror (result_code));
    printf ("Response code: %ld\n", response_code);
    curl_easy_cleanup (curl);
    curl_slist_free_all(list);

    if (response_data) {
        /* TODO is a simple "UUID COMMAND" syntax sufficient, or should we
         * return BSON/JSON from the server, like Ubuntu One does? */
        split_string (response_data, &command);
        printf ("Got UUID: %s\n", response_data);
        printf ("Got command: %s\n", command);
        free (response_data);
    }

    return result_code == CURLE_OK && response_code == 200;
}

void
destroy_key_and_value (gpointer key, gpointer value, gpointer user_data)
{
    if (key)
        g_free (key);
    if (value)
        g_free (value);
}

GHashTable*
parse_report (const char* report_path)
{
    /* We'll eventually modify the contents of the report, rather than sending
     * it as-is, to make it more amenable to what the server has to stick in
     * the database, and thus creating less work server-side.
     */

    GMappedFile* fp = NULL;
    GHashTable* hash_table = NULL;
    gchar* contents = NULL;
    gsize file_len = 0;
    /* Our position in the file. */
    size_t p = 0;
    /* The end or length of the token. */
    size_t token_p = 0;
    char* key = NULL;
    char* value = NULL;
    size_t value_p = 0;
    GError* err = NULL;

    /* TODO handle the file being modified underneath us. */
    fp = g_mapped_file_new (report_path, FALSE, &err);
    if (err) {
        g_error ("Unable to map report: %s\n", err->message);
        g_error_free (err);
        return NULL;
    }
        
    contents = g_mapped_file_get_contents (fp);
    file_len = g_mapped_file_get_length (fp);
    hash_table = g_hash_table_new (g_str_hash, g_str_equal);

    while (p < file_len) {
        assert (p == 0 || contents[p - 1] == '\n');
        if (contents[p] == ' ') {
            /* Skip the space. */
            p++;
            token_p = p;
            while (token_p < file_len && contents[token_p] != '\n') token_p++;
            token_p = token_p - p;
            /* Space for the leading newline too. */
            value = realloc (value, 1 + value_p + token_p + 1);
            value[value_p] = '\n';
            /* Again, skip past the leading newline. */
            value_p += 1;
            memcpy (value + value_p, contents + p, token_p);
            /* We don't extend past the null terminator, because we want to
             * overwrite it on the next pass. */
            value_p += token_p;
            value[value_p] = '\0';
            p += token_p + 1;
            g_hash_table_insert (hash_table, key, value);
        } else {
            /* Reset the value pointer. */
            value_p = 0;
            /* Key. */
            token_p = p;
            while (token_p < file_len) {
                if (contents[token_p] != ':' && contents[token_p] != '\n')
                    token_p++;
                else
                    break;
            }
            token_p = token_p - p;
            /* When we load the whole file into memory, we wont need to malloc
             * anymore, as we can add null-terminators in the existing string
             * buffer. */
            key = malloc (token_p + 1);
            memcpy (key, contents + p, token_p);
            key[token_p] = '\0';
            p += token_p + 1;

            token_p = p;
            while (token_p < file_len && contents[token_p] != '\n') token_p++;
            /* Skip any leading spaces. */
            while (p < file_len && contents[p] == ' ') p++;
            token_p = token_p - p;
            if (token_p == 0) {
                /* Empty value. The key likely has a child. */
                value = NULL;
            } else {
                /* Value. */
                value = malloc (token_p + 1);
                memcpy (value, contents + p, token_p);
                value[token_p] = '\0';
            }
            p += token_p + 1;
            g_hash_table_insert (hash_table, key, value);
        }
    }
    g_mapped_file_free (fp);
    return hash_table;
}

gboolean
parse_and_upload_report (const char* crash_file)
{
    GHashTable* report = NULL;
    gboolean success = FALSE;
    int message_len = 0;
    const char* message_data = NULL;
    bson b[1];

    report = parse_report (crash_file);
    if (!report) {
        g_warning ("Unable to parse report: %s\n", crash_file);
        return FALSE;
    }

    bsonify (report, b, &message_data, &message_len);
    /* TODO hold onto report['CoreDump'], upload as-is for now, unbase64ing on
     * the other side, just to get something in place to work on the retracers.
     * */
    g_hash_table_foreach (report, destroy_key_and_value, NULL);
    g_hash_table_destroy (report);
    success = upload_report (message_data, message_len);
    bson_destroy (b);

    return success;
}

void
create_file (const char* upload)
{
    char* upload_file = g_strdup (upload);
    char* crash_file = upload_to_crash_file (upload_file);

    if (g_file_test (crash_file, G_FILE_TEST_EXISTS)) {
        g_message ("%s exists", crash_file);
        if (online_state && parse_and_upload_report (crash_file)) {
            if (g_unlink (upload_file))
                g_warning ("Unable to remove: %s", upload_file);
            free (crash_file);
        } else {
            g_warning ("Adding to queue: %s", upload_file);
            g_queue_push_head (report_queue, (gpointer)upload_file);
        }
    } else {
        /* Already cleaned up? Nothing more we can do. */
        if (g_unlink (upload_file))
            g_warning ("Unable to remove: %s", upload_file);
        free (crash_file);
    }
    g_free (upload_file);
}

gboolean
remove_from_report_queue (const char* upload_file) {
    GList* queued_crash = NULL;
    queued_crash = g_queue_find_custom (report_queue, upload_file,
                                        (GCompareFunc)strcmp);
    if (queued_crash) {
        gpointer data = queued_crash->data;
        if (!g_queue_remove (report_queue, queued_crash->data))
            g_warning ("Unable to remove from queue: %s", upload_file);
        g_free (data);
        g_list_free (queued_crash);
        return TRUE;
    } else {
        return FALSE;
    }
}

void
delete_file (const char* upload_file)
{
    g_message ("deleted: %s", upload_file);
    if (remove_from_report_queue (upload_file))
        g_warning ("%s deleted before being processed.", upload_file);
}

void
changed_event (GFileMonitor* monitor, GFile *file, GFile *other_file,
               GFileMonitorEvent event_type, gpointer user_data)
{
    char* path = NULL;
    const char *p, *ext = NULL;

    path = g_file_get_path (file);
    /* Find the file extension. */
    ext = strrchr(path, '.');
    if (ext == NULL || strcmp(++ext, "upload") != 0)
        return;
    if (event_type == G_FILE_MONITOR_EVENT_CREATED)
        create_file(path);
    else if (event_type == G_FILE_MONITOR_EVENT_DELETED)
        delete_file(path);

    g_free(path);
}

gboolean
process_queue (void) {
    g_warning ("Processing queue.");
    GList* list = NULL;
    char *upload_file, *crash_file = NULL;
    list = report_queue->head;
    while (list) {
        GList* next = list->next;
        upload_file = list->data;
        if (g_file_test (upload_file, G_FILE_TEST_EXISTS)) {
            crash_file = upload_to_crash_file (upload_file);
            if (online_state && parse_and_upload_report (crash_file)) {
                if (g_unlink (upload_file))
                    g_warning ("Unable to remove: %s", upload_file);
                remove_from_report_queue (upload_file);
            }
            free (crash_file);
        } else {
            remove_from_report_queue (upload_file);
        }
        list = next;
    }
    return TRUE;
}

void
process_existing_files (void)
{
    GDir* dir = NULL;
    const gchar *file, *ext = NULL;
    gchar* upload_file = NULL;
    char* crash_file = NULL;

    dir = g_dir_open ("/var/crash", 0, NULL);
    while ((file = g_dir_read_name (dir)) != NULL) {
        upload_file = g_build_filename ("/var/crash", file, NULL);
        ext = strrchr(upload_file, '.');
        if (strcmp(++ext, "upload") == 0) {
            crash_file = upload_to_crash_file (upload_file);
            if (online_state && parse_and_upload_report (crash_file)) {
                if (g_unlink (upload_file))
                    g_warning ("Unable to remove: %s", upload_file);
                free (crash_file);
                free (upload_file);
            } else {
                g_queue_push_head (report_queue, upload_file);
            }
        } else {
            free (upload_file);
        }
    }
    g_dir_close (dir);
}

void
network_manager_state_changed (GDBusConnection* connection,
                               const gchar* sender_name, const gchar*
                               object_path, const gchar* interface_name,
                               const gchar* signal_name, GVariant* parameters,
                               gpointer user_data)
{
    guint32 res;
    if (parameters) {
        g_variant_get_child (parameters, 0, "u", &res);
        online_state = NM_STATE_CONNECTED_GLOBAL == res;
        if (online_state)
            process_queue ();
    }
}

gboolean
monitor_network_manager_connectivity (void)
{
    /* TODO When GNetworkMonitor lands in the archive, use that if detected. */
    GDBusConnection* system_bus = NULL;
    GError* err = NULL;
    GVariant* result = NULL;
    GVariant* res = NULL;
    system_bus = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &err);
    if (err)
        goto error;

    g_dbus_connection_signal_subscribe (system_bus,
                                        NETWORK_MANAGER_SERVICE,
                                        NETWORK_MANAGER_INTERFACE,
                                        "StateChanged",
                                        NULL,
                                        NULL,
                                        (GDBusSignalFlags) NULL,
                                        network_manager_state_changed,
                                        NULL,
                                        NULL);

    if (err)
        goto error;
    result = g_dbus_connection_call_sync (system_bus,
                                          NETWORK_MANAGER_SERVICE,
                                          NETWORK_MANAGER_OBJECT,
                                          DBUS_PROPERTIES_INTERFACE,
                                          "Get",
                                          g_variant_new ("(ss)",
                                                     NETWORK_MANAGER_INTERFACE,
                                                     "state"),
                                          NULL,
                                          G_DBUS_CALL_FLAGS_NONE,
                                          3000,
                                          NULL,
                                          &err);
    if (err)
        goto error;

    if (result) {
        g_variant_get (result, "(v)", &res);
        online_state = NM_STATE_CONNECTED_GLOBAL == g_variant_get_uint32 (res);
        if (res)
            g_variant_unref (res);
        if (result)
            g_variant_unref (result);
    }

    g_object_unref (system_bus);
    return TRUE;

    error:
        g_warning ("Could not monitor Network Manager: %s", err->message);
        g_error_free (err);
        g_object_unref (system_bus);
        return FALSE;
}

void daemonize (void)
{
    pid_t pid, sid;

    pid = fork();
    if (pid < 0)
        exit (EXIT_FAILURE);
    if (pid > 0)
        exit (EXIT_SUCCESS);
    umask (0);
    sid = setsid ();
    if (sid < 0)
        exit (EXIT_FAILURE);

    if ((chdir ("/")) < 0)
        exit (EXIT_FAILURE);

    close (STDIN_FILENO);
    close (STDOUT_FILENO);
    close (STDERR_FILENO);
}

void
exit_if_already_running (void)
{
    int lock_fd = 0;
    int rc = 0;
    lock_fd = open ("/tmp/.reporterd-lock", O_CREAT | O_RDWR, 0666);
    rc = flock (lock_fd, LOCK_EX | LOCK_NB);
    if (rc) {
        if (EWOULDBLOCK == errno) {
            g_warning ("Another instance is already running.");
            exit (1);
        } else {
            g_warning ("Could not create lock file: %s", strerror (errno));
        }
    }
}

static void
handle_signals (int signo)
{
    if (loop)
        g_main_loop_quit (loop);
    else
        exit (0);
}

static void
setup_signals (void)
{
    struct sigaction action;
    sigset_t mask;

    sigemptyset (&mask);
    action.sa_handler = handle_signals;
    action.sa_mask = mask;
    action.sa_flags = 0;
    sigaction (SIGTERM, &action, NULL);
    sigaction (SIGINT, &action, NULL);
}

int
main (int argc, char** argv)
{
    GFile* path = NULL;
    GFileMonitor* monitor = NULL;
    GError* err = NULL;
    gulong handler = 0;

    setup_signals ();
    exit_if_already_running ();
    parse_arguments (&argc, &argv);
    if (run_daemonized)
        daemonize ();

    g_type_init ();

    report_queue = g_queue_new ();
    mkdir ("/var/crash", 1777);
    path = g_file_new_for_path ("/var/crash");
    monitor = g_file_monitor_directory (path, G_FILE_MONITOR_NONE, NULL, &err);
    g_object_unref (path);
    if (err) {
        g_warning ("Unable to monitor /var/crash: %s\n", err->message);
        g_error_free (err);
    } else {
        handler = g_signal_connect (monitor, "changed",
                                    G_CALLBACK (changed_event), NULL);
    }
    g_timeout_add_seconds (60 * 5, (GSourceFunc) process_queue, NULL);

    monitor_network_manager_connectivity ();
    process_existing_files ();
    loop = g_main_loop_new (NULL, FALSE);
    g_main_loop_run (loop);

    g_signal_handler_disconnect (monitor, handler);
    g_object_unref (monitor);
    g_queue_foreach (report_queue, (GFunc)g_free, NULL);
    g_queue_free (report_queue);
	return 0;
}
